# SPDX-License-Identifier: MIT

import os
import os.path
import platform
import shutil
import stat
import subprocess
import sys
import sysconfig
import tempfile

import pytest

from filelock import FileLock


def _build_and_reinstall_build(test_mode):
    temp = tempfile.mkdtemp()
    try:
        subprocess.check_output(
            [sys.executable, '-m', 'build', '--{}'.format(test_mode), '--no-isolation', '--outdir', temp],
        )
        dist_file = next(d for d in os.listdir(temp) if d.endswith('.whl' if test_mode == 'wheel' else '.tar.gz'))
        subprocess.check_call(
            [
                sys.executable,
                '-m',
                'pip',
                'install',
                '--upgrade',  # ``--upgrade`` will uninstall build prior to installing the ``dist_file``
                os.path.join(temp, dist_file),
            ],
        )
    finally:
        shutil.rmtree(temp)


def _one_time_setup():
    test_mode = os.environ.get('TEST_MODE')
    if not test_mode:
        return

    if test_mode == 'path':
        project_root = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
        sys.path.insert(0, os.path.join(project_root, 'src'))
    elif test_mode in {'sdist', 'wheel'}:
        status_marker_file = os.path.join(os.environ['TEST_STATUS_DIR'], 'status-marker')
        with FileLock(status_marker_file + '.lock'):
            if not os.path.exists(status_marker_file):
                _build_and_reinstall_build(test_mode)

                with open(status_marker_file, 'wb'):
                    pass


_one_time_setup()


def pytest_addoption(parser):
    os.environ['PYTHONWARNINGS'] = 'ignore:DEPRECATION::pip._internal.cli.base_command'  # for when not run within tox
    os.environ['PIP_DISABLE_PIP_VERSION_CHECK'] = '1'  # do not pollute stderr with upgrade advisory
    parser.addoption('--run-integration', action='store_true', help='run the integration tests')
    parser.addoption('--only-integration', action='store_true', help='only run the integration tests')

    # pytest-xdist does not work on Windows PyPy2 - we need to disable it
    if platform.python_implementation() == 'PyPy' and sys.version_info[0] == 2 and os.name == 'nt':
        # the patch here is compatible with pytest 4 + pytest-xdist 1.31 (last release on python2)
        num_process_option_xdist = next(o for o in parser.getgroup('xdist').options if o.dest == 'numprocesses')
        # change the type to return None -> no parallel runs
        num_process_option_xdist._attrs['type'] = lambda s: None  # noqa


PYPY3_WIN_VENV_BAD = platform.python_implementation() == 'PyPy' and sys.version_info[0] == 3 and os.name == 'nt'
PYPY3_WIN_M = 'https://foss.heptapod.net/pypy/pypy/-/issues/3323 and https://foss.heptapod.net/pypy/pypy/-/issues/3321'


def pytest_collection_modifyitems(config, items):
    skip_int = pytest.mark.skip(reason='integration tests not run (no --run-integration flag)')
    skip_other = pytest.mark.skip(reason='only integration tests are run (got --only-integration flag)')

    if config.getoption('--run-integration') and config.getoption('--only-integration'):  # pragma: no cover
        raise pytest.UsageError("--run-integration and --only-integration can't be used together, choose one")

    if len(items) == 1:  # do not require flags if called directly
        return
    for item in items:
        is_integration_file = is_integration(item)
        if PYPY3_WIN_VENV_BAD and item.get_closest_marker('isolated'):
            if not (is_integration_file and item.originalname == 'test_build') or (
                hasattr(item, 'callspec') and '--no-isolation' not in item.callspec.params.get('args', [])
            ):
                item.add_marker(pytest.mark.xfail(reason=PYPY3_WIN_M, strict=True))
        if is_integration_file:  # pragma: no cover
            if not config.getoption('--run-integration') and not config.getoption('--only-integration'):
                item.add_marker(skip_int)
        elif config.getoption('--only-integration'):  # pragma: no cover
            item.add_marker(skip_other)
    # run integration tests after unit tests
    items.sort(key=lambda i: 1 if is_integration(i) else 0)


def is_integration(item):
    return os.path.basename(item.location[0]) == 'test_integration.py'


@pytest.fixture(scope='session', autouse=True)
def ensure_syconfig_vars_created():
    # the config vars are globally cached and may use get_path, make sure they are created
    sysconfig.get_config_vars()


@pytest.fixture
def packages_path():
    return os.path.realpath(os.path.join(__file__, '..', 'packages'))


@pytest.fixture
def legacy_path(packages_path):
    return os.path.join(packages_path, 'legacy')


@pytest.fixture
def test_flit_path(packages_path):
    return os.path.join(packages_path, 'test-flit')


@pytest.fixture
def test_bad_syntax_path(packages_path):
    return os.path.join(packages_path, 'test-bad-syntax')


@pytest.fixture
def test_no_backend_path(packages_path):
    return os.path.join(packages_path, 'test-no-backend')


@pytest.fixture
def test_no_requires_path(packages_path):
    return os.path.join(packages_path, 'test-no-requires')


@pytest.fixture
def test_optional_hooks_path(packages_path):
    return os.path.join(packages_path, 'test-optional-hooks')


@pytest.fixture
def test_typo(packages_path):
    return os.path.join(packages_path, 'test-typo')


@pytest.fixture
def test_no_permission(packages_path):
    path = os.path.join(packages_path, 'test-no-permission')
    file = os.path.join(path, 'pyproject.toml')
    orig_stat = os.stat(file).st_mode

    os.chmod(file, ~stat.S_IRWXU)

    yield os.path.join(packages_path, 'test-no-permission')

    os.chmod(file, orig_stat)


@pytest.fixture
def tmp_dir():
    path = tempfile.mkdtemp(prefix='python-build-test-')

    yield path

    shutil.rmtree(path)



# SPDX-License-Identifier: MIT

import contextlib
import io
import os
import sys

import pytest

import build
import build.__main__


if sys.version_info >= (3,):  # pragma: no cover
    build_open_owner = 'builtins'
else:  # pragma: no cover
    build_open_owner = 'build'


cwd = os.getcwd()
out = os.path.join(cwd, 'dist')


@pytest.mark.parametrize(
    ('cli_args', 'build_args'),
    [
        (
            [],
            [cwd, out, ['sdist', 'wheel'], {}, True, False],
        ),
        (
            ['-n'],
            [cwd, out, ['sdist', 'wheel'], {}, False, False],
        ),
        (
            ['-s'],
            [cwd, out, ['sdist'], {}, True, False],
        ),
        (
            ['-w'],
            [cwd, out, ['wheel'], {}, True, False],
        ),
        (
            ['source'],
            ['source', os.path.join('source', 'dist'), ['sdist', 'wheel'], {}, True, False],
        ),
        (
            ['-o', 'out'],
            [cwd, 'out', ['sdist', 'wheel'], {}, True, False],
        ),
        (
            ['source', '-o', 'out'],
            ['source', 'out', ['sdist', 'wheel'], {}, True, False],
        ),
        (
            ['-x'],
            [cwd, out, ['sdist', 'wheel'], {}, True, True],
        ),
        (
            ['-C--flag1', '-C--flag2'],
            [cwd, out, ['sdist', 'wheel'], {'--flag1': '', '--flag2': ''}, True, False],
        ),
        (
            ['-C--flag=value'],
            [cwd, out, ['sdist', 'wheel'], {'--flag': 'value'}, True, False],
        ),
        (
            ['-C--flag1=value', '-C--flag2=other_value', '-C--flag2=extra_value'],
            [cwd, out, ['sdist', 'wheel'], {'--flag1': 'value', '--flag2': ['other_value', 'extra_value']}, True, False],
        ),
    ],
)
def test_parse_args(mocker, cli_args, build_args):
    mocker.patch('build.__main__.build_package')

    build.__main__.main(cli_args)
    build.__main__.build_package.assert_called_with(*build_args)


def test_prog():
    out = io.StringIO()

    if sys.version_info >= (3,):  # pragma: no cover
        with pytest.raises(SystemExit):
            with contextlib.redirect_stdout(out):
                build.__main__.main(['--help'], prog='something')

        assert out.getvalue().startswith('usage: something [-h]')
    else:  # pragma: no cover
        with pytest.raises(SystemExit):
            build.__main__.main(['--help'], prog='something')


def test_version(capsys):
    with pytest.raises(SystemExit):
        build.__main__.main(['--version'])
    out, err = capsys.readouterr()
    target = out if sys.version_info[0] == 3 else err
    assert target.startswith('build {}'.format(build.__version__))


@pytest.mark.isolated
def test_build_isolated(mocker, test_flit_path):
    build_cmd = mocker.patch('build.ProjectBuilder.build')
    required_cmd = mocker.patch(
        'build.ProjectBuilder.get_dependencies',
        side_effect=[
            ['dep1', 'dep2'],
        ],
    )
    mocker.patch('build.__main__._error')
    install = mocker.patch('build.env._IsolatedEnvVenvPip.install')

    build.__main__.build_package(test_flit_path, '.', ['sdist'])

    install.assert_any_call({'flit_core >=2,<3'})

    required_cmd.assert_called_with('sdist')
    install.assert_any_call(['dep1', 'dep2'])

    build_cmd.assert_called_with('sdist', '.', {})


def test_build_no_isolation_check_deps_empty(mocker, test_flit_path):
    # check_dependencies = []
    build_cmd = mocker.patch('build.ProjectBuilder.build')
    mocker.patch('build.ProjectBuilder.check_dependencies', return_value=[])

    build.__main__.build_package(test_flit_path, '.', ['sdist'], isolation=False)

    build_cmd.assert_called_with('sdist', '.', {})


@pytest.mark.parametrize(
    ['missing_deps', 'output'],
    [
        ([('foo',)], '\n\tfoo'),
        ([('foo',), ('bar', 'baz', 'qux')], '\n\tfoo\n\tbar\n\tbaz -> qux'),
    ],
)
def test_build_no_isolation_with_check_deps(mocker, test_flit_path, missing_deps, output):
    error = mocker.patch('build.__main__._error')
    build_cmd = mocker.patch('build.ProjectBuilder.build')
    mocker.patch('build.ProjectBuilder.check_dependencies', return_value=missing_deps)

    build.__main__.build_package(test_flit_path, '.', ['sdist'], isolation=False)

    build_cmd.assert_called_with('sdist', '.', {})
    error.assert_called_with('Missing dependencies:' + output)


@pytest.mark.isolated
def test_build_raises_build_exception(mocker, test_flit_path):
    error = mocker.patch('build.__main__._error')
    mocker.patch('build.ProjectBuilder.get_dependencies', side_effect=build.BuildException)
    mocker.patch('build.env._IsolatedEnvVenvPip.install')

    build.__main__.build_package(test_flit_path, '.', ['sdist'])

    error.assert_called_with('')


@pytest.mark.isolated
def test_build_raises_build_backend_exception(mocker, test_flit_path):
    error = mocker.patch('build.__main__._error')
    mocker.patch('build.ProjectBuilder.get_dependencies', side_effect=build.BuildBackendException(Exception('a')))
    mocker.patch('build.env._IsolatedEnvVenvPip.install')

    build.__main__.build_package(test_flit_path, '.', ['sdist'])
    msg = "Backend operation failed: Exception('a'{})".format(',' if sys.version_info < (3, 7) else '')
    error.assert_called_with(msg)



"""
Creates and manages isolated build environments.
"""
import abc
import os
import platform
import shutil
import subprocess
import sys
import sysconfig
import tempfile

from types import TracebackType
from typing import Iterable, Optional, Tuple, Type

import packaging.requirements
import packaging.version

import build

from ._compat import abstractproperty, add_metaclass


if sys.version_info < (3, 8):
    import importlib_metadata as metadata
else:
    from importlib import metadata

try:
    import virtualenv
except ImportError:
    virtualenv = None


@add_metaclass(abc.ABCMeta)
class IsolatedEnv(object):
    """Abstract base of isolated build environments, as required by the build project."""

    @abstractproperty
    def executable(self):  # type: () -> str
        """The executable of the isolated build environment."""
        raise NotImplementedError

    @abstractproperty
    def scripts_dir(self):  # type: () -> str
        """The scripts directory of the isolated build environment."""
        raise NotImplementedError

    @abc.abstractmethod
    def install(self, requirements):  # type: (Iterable[str]) -> None
        """
        Install packages from PEP 508 requirements in the isolated build environment.

        :param requirements: PEP 508 requirements
        """
        raise NotImplementedError


class IsolatedEnvBuilder(object):
    """Builder object for isolated environments."""

    _has_virtualenv = None  # type: Optional[bool]

    def __init__(self):  # type: () -> None
        self._path = None  # type: Optional[str]

    def _should_use_virtualenv(self):  # type: () -> Optional[bool]
        # virtualenv might be incompatible if it was installed separately
        # from build. This verifies that virtualenv and all of its
        # dependencies are installed as specified by build.
        if self._has_virtualenv is None:
            self.__class__._has_virtualenv = virtualenv is not None and not any(
                packaging.requirements.Requirement(d[1]).name == 'virtualenv'
                for d in build.check_dependency('build[virtualenv]')
                if len(d) > 1
            )
        return self._has_virtualenv

    def __enter__(self):  # type: () -> IsolatedEnv
        """
        Create an isolated build environment.

        :return: The isolated build environment
        """
        self._path = tempfile.mkdtemp(prefix='build-env-')
        try:
            # use virtualenv on Python 2 or when valid virtualenv is available (as it's faster than venv)
            if sys.version_info < (3,) or self._should_use_virtualenv():
                executable, scripts_dir = _create_isolated_env_virtualenv(self._path)
            else:
                executable, scripts_dir = _create_isolated_env_venv(self._path)
            return _IsolatedEnvVenvPip(path=self._path, python_executable=executable, scripts_dir=scripts_dir)
        except Exception:  # cleanup folder if creation fails
            self.__exit__(*sys.exc_info())
            raise

    def __exit__(self, exc_type, exc_val, exc_tb):
        # type: (Optional[Type[BaseException]], Optional[BaseException], Optional[TracebackType]) -> None
        """
        Delete the created isolated build environment.

        :param exc_type: The type of exception raised (if any)
        :param exc_val: The value of exception raised (if any)
        :param exc_tb: The traceback of exception raised (if any)
        """
        if self._path is not None and os.path.exists(self._path):  # in case the user already deleted skip remove
            shutil.rmtree(self._path)


class _IsolatedEnvVenvPip(IsolatedEnv):
    """
    Isolated build environment context manager

    Non-standard paths injected directly to sys.path will still be passed to the environment.
    """

    def __init__(self, path, python_executable, scripts_dir):
        # type: (str, str, str) -> None
        """
        :param path: The path where the environment exists
        :param python_executable: The python executable within the environment
        """
        self._path = path
        self._python_executable = python_executable
        self._scripts_dir = scripts_dir

    @property
    def path(self):  # type: () -> str
        """The location of the isolated build environment."""
        return self._path

    @property
    def executable(self):  # type: () -> str
        """The python executable of the isolated build environment."""
        return self._python_executable

    @property
    def scripts_dir(self):  # type: () -> str
        return self._scripts_dir

    def install(self, requirements):  # type: (Iterable[str]) -> None
        """
        Install packages from PEP 508 requirements in the isolated build environment.

        :param requirements: PEP 508 requirement specification to install

        :note: Passing non-PEP 508 strings will result in undefined behavior, you *should not* rely on it. It is
               merely an implementation detail, it may change any time without warning.
        """
        if not requirements:
            return

        # pip does not honour environment markers in command line arguments
        # but it does for requirements from a file
        with tempfile.NamedTemporaryFile('w+', prefix='build-reqs-', suffix='.txt', delete=False) as req_file:
            req_file.write(os.linesep.join(requirements))
        try:
            cmd = [
                self.executable,
                '-{}m'.format('E' if sys.version_info[0] == 2 else 'I'),
                'pip',
                'install',
                '--use-pep517',
                '--no-warn-script-location',
                '-r',
                os.path.abspath(req_file.name),
            ]
            subprocess.check_call(cmd)
        finally:
            os.unlink(req_file.name)


def _create_isolated_env_virtualenv(path):  # type: (str) -> Tuple[str, str]
    """
    On Python 2 we use the virtualenv package to provision a virtual environment.

    :param path: The path where to create the isolated build environment
    :return: The Python executable and script folder
    """
    cmd = [str(path), '--no-setuptools', '--no-wheel', '--activators', '']
    result = virtualenv.cli_run(cmd, setup_logging=False)
    executable = str(result.creator.exe)
    script_dir = str(result.creator.script_dir)
    return executable, script_dir


# venv only exists on Python 3+
if sys.version_info >= (3,):  # noqa: C901

    def _create_isolated_env_venv(path):  # type: (str) -> Tuple[str, str]
        """
        On Python 3 we use the venv package from the standard library.

        :param path: The path where to create the isolated build environment
        :return: The Python executable and script folder
        """
        import venv

        venv.EnvBuilder(with_pip=True).create(path)
        executable, script_dir, purelib = _find_executable_and_scripts(path)

        # Get the version of pip in the environment
        pip_distribution = next(iter(metadata.distributions(name='pip', path=[purelib])))
        pip_version = packaging.version.Version(pip_distribution.version)

        needs_pip_upgrade = False
        if pip_version < packaging.version.Version('19.1'):
            # PEP-517 and manylinux1 was first implemented in 19.1
            needs_pip_upgrade = True
        elif platform.system() == 'Darwin':
            # macOS 11+ name scheme change requires 20.3. Intel macOS 11.0 can be told to report 10.16 for backwards
            # compatibility; but that also fixes earlier versions of pip so this is only needed for 11+.

            if int(platform.mac_ver()[0].split('.')[0]) >= 11:
                needs_pip_upgrade = pip_version < packaging.version.Version('20.3') or (
                    pip_version < packaging.version.Version('21.0.1')  # Apple Silicon macOS requires 21.0.1+
                    and sys.version_info >= (3, 6)
                    and platform.machine() != 'x86_64'
                )
        if needs_pip_upgrade:
            subprocess.check_call([executable, '-m', 'pip', 'install', '-U', 'pip'])

        # Avoid the setuptools from ensurepip to break the isolation
        subprocess.check_call([executable, '-m', 'pip', 'uninstall', 'setuptools', '-y'])
        return executable, script_dir

    def _find_executable_and_scripts(path):  # type: (str) -> Tuple[str, str, str]
        """
        Detect the Python executable and script folder of a virtual environment.

        :param path: The location of the virtual environment
        :return: The Python executable, script folder, and purelib folder
        """
        config_vars = sysconfig.get_config_vars().copy()  # globally cached, copy before altering it
        config_vars['base'] = path
        env_scripts = sysconfig.get_path('scripts', vars=config_vars)
        if not env_scripts:
            raise RuntimeError("Couldn't get environment scripts path")
        exe = 'pypy3' if platform.python_implementation() == 'PyPy' else 'python'
        if os.name == 'nt':
            exe = '{}.exe'.format(exe)
        executable = os.path.join(env_scripts, exe)
        if not os.path.exists(executable):
            raise RuntimeError('Virtual environment creation failed, executable {} missing'.format(executable))

        purelib = sysconfig.get_path('purelib', vars=config_vars)
        if not purelib:
            raise RuntimeError("Couldn't get environment purelib folder")
        return executable, env_scripts, purelib


__all__ = (
    'IsolatedEnvBuilder',
    'IsolatedEnv',
)



import os
import sys

from textwrap import dedent
from zipfile import ZipFile


name = 'demo_pkg_inline'
pkg_name = name.replace('_', '-')

version = '1.0.0'
dist_info = '{}-{}.dist-info'.format(name, version)

metadata = '{}/METADATA'.format(dist_info)
wheel = '{}/WHEEL'.format(dist_info)
entry_points = '{}/entry_points.txt'.format(dist_info)
record = '{}/RECORD'.format(dist_info)
init = '{}/__init__.py'.format(name)
content = {
    init: "def do():\n    print('greetings from {}')".format(name),
    metadata: """
        Metadata-Version: 2.1
        Name: {}
        Version: {}
        Summary: Summary of package
        Home-page: Does not exists
        Author: someone
        Author-email: a@o.com
        License: MIT
        Platform: ANY

        Desc
       """.format(
        pkg_name, version
    ),
    wheel: """
        Wheel-Version: 1.0
        Generator: {}-{}
        Root-Is-Purelib: true
        Tag: py3-none-any
       """.format(
        name, version
    ),
    '{}/top_level.txt'.format(dist_info): name,
    entry_points: '\n[console_scripts]\ndemo-pkg-inline = demo_pkg_inline:do',
    record: """
        {0}/__init__.py,,
        {1}/METADATA,,
        {1}/WHEEL,,
        {1}/top_level.txt,,
        {1}/RECORD,,
       """.format(
        name, dist_info
    ),
}


def build_wheel(wheel_directory, metadata_directory=None, config_settings=None):
    base_name = '{}-{}-py{}-none-any.whl'.format(name, version, sys.version_info.major)
    path = os.path.join(wheel_directory, base_name)
    with ZipFile(str(path), 'w') as zip_file_handler:
        for arc_name, data in content.items():
            zip_file_handler.writestr(arc_name, dedent(data).strip())
    print('created wheel {}'.format(path))
    return base_name


def get_requires_for_build_wheel(config_settings):
    return []



# SPDX-License-Identifier: MIT
import collections
import os
import platform
import shutil
import subprocess
import sys
import sysconfig

import pytest

from packaging.version import Version

import build.env


IS_PYPY3 = sys.version_info[0] == 3 and platform.python_implementation() == 'PyPy'
IS_PY35 = sys.version_info[:2] == (3, 5)
IS_PY2 = sys.version_info[0] == 2


@pytest.mark.isolated
def test_isolation():
    subprocess.check_call([sys.executable, '-c', 'import build.env'])
    with build.env.IsolatedEnvBuilder() as env:
        with pytest.raises(subprocess.CalledProcessError):
            debug = 'import sys; import os; print(os.linesep.join(sys.path));'
            subprocess.check_call([env.executable, '-c', '{} import build.env'.format(debug)])


@pytest.mark.isolated
def test_isolated_environment_install(mocker):
    with build.env.IsolatedEnvBuilder() as env:
        mocker.patch('subprocess.check_call')

        env.install([])
        subprocess.check_call.assert_not_called()

        env.install(['some', 'requirements'])
        if not IS_PY35:
            subprocess.check_call.assert_called()
        args = subprocess.check_call.call_args[0][0][:-1]
        assert args == [
            env.executable,
            '-{}m'.format('E' if IS_PY2 else 'I'),
            'pip',
            'install',
            '--use-pep517',
            '--no-warn-script-location',
            '-r',
        ]


@pytest.mark.skipif(IS_PY2, reason='venv module used on Python 3 only')
@pytest.mark.skipif(IS_PYPY3, reason='PyPy3 uses get path to create and provision venv')
def test_fail_to_get_script_path(mocker):
    mocker.patch.object(build.env, 'virtualenv', None)
    get_path = mocker.patch('sysconfig.get_path', return_value=None)
    with pytest.raises(RuntimeError, match="Couldn't get environment scripts path"):
        env = build.env.IsolatedEnvBuilder()
        with env:
            pass
    assert not os.path.exists(env._path)
    assert get_path.call_count == 1


@pytest.mark.skipif(IS_PY2, reason='venv module used on Python 3 only')
@pytest.mark.skipif(IS_PYPY3, reason='PyPy3 uses get path to create and provision venv')
def test_fail_to_get_purepath(mocker):
    mocker.patch.object(build.env, 'virtualenv', None)
    sysconfig_get_path = sysconfig.get_path
    mocker.patch(
        'sysconfig.get_path',
        side_effect=lambda path, *args, **kwargs: '' if path == 'purelib' else sysconfig_get_path(path, *args, **kwargs),
    )

    with pytest.raises(RuntimeError, match="Couldn't get environment purelib folder"):
        with build.env.IsolatedEnvBuilder():
            pass


@pytest.mark.skipif(IS_PY2, reason='venv module used on Python 3 only')
@pytest.mark.skipif(IS_PYPY3, reason='PyPy3 uses get path to create and provision venv')
def test_executable_missing_post_creation(mocker):
    mocker.patch.object(build.env, 'virtualenv', None)
    original_get_path = sysconfig.get_path

    def _get_path(name, vars):  # noqa
        shutil.rmtree(vars['base'])
        return original_get_path(name, vars=vars)

    get_path = mocker.patch('sysconfig.get_path', side_effect=_get_path)
    with pytest.raises(RuntimeError, match='Virtual environment creation failed, executable .* missing'):
        with build.env.IsolatedEnvBuilder():
            pass
    assert get_path.call_count == 1


def test_isolated_env_abstract():
    with pytest.raises(TypeError):
        build.env.IsolatedEnv()


def test_isolated_env_has_executable_still_abstract():
    class Env(build.env.IsolatedEnv):  # noqa
        @property
        def executable(self):
            raise NotImplementedError

    with pytest.raises(TypeError):
        Env()


def test_isolated_env_has_install_still_abstract():
    class Env(build.env.IsolatedEnv):  # noqa
        def install(self, requirements):
            raise NotImplementedError

    with pytest.raises(TypeError):
        Env()


@pytest.mark.isolated
def test_default_pip_is_never_too_old():
    with build.env.IsolatedEnvBuilder() as env:
        version = subprocess.check_output(
            [env.executable, '-c', 'import pip; print(pip.__version__)'], universal_newlines=True
        ).strip()
        assert Version(version) >= Version('19.1')


@pytest.mark.isolated
@pytest.mark.parametrize('pip_version', ['20.2.0', '20.3.0', '21.0.0', '21.0.1'])
@pytest.mark.parametrize('arch', ['x86_64', 'arm64'])
@pytest.mark.skipif(IS_PY35, reason="Python 3.5 does not run on macOS 11, and pip can't upgrade to 21 there")
@pytest.mark.skipif(IS_PY2, reason='venv module used on Python 3 only')
def test_pip_needs_upgrade_mac_os_11(mocker, pip_version, arch):
    SimpleNamespace = collections.namedtuple('SimpleNamespace', 'version')

    check_call = mocker.patch('subprocess.check_call')
    mocker.patch('platform.system', return_value='Darwin')
    mocker.patch('platform.machine', return_value=arch)
    mocker.patch('platform.mac_ver', return_value=('11.0', ('', '', ''), ''))
    mocker.patch('build.env.metadata.distributions', return_value=(SimpleNamespace(version=pip_version),))
    mocker.patch.object(build.env, 'virtualenv', None)  # hide virtualenv

    min_version = Version('20.3' if arch == 'x86_64' else '21.0.1')
    with build.env.IsolatedEnvBuilder():
        if Version(pip_version) < min_version:
            upgrade_call, uninstall_call = check_call.call_args_list
            assert upgrade_call[0][0][1:] == ['-m', 'pip', 'install', '-U', 'pip']
            assert uninstall_call[0][0][1:] == ['-m', 'pip', 'uninstall', 'setuptools', '-y']
        else:
            (uninstall_call,) = check_call.call_args_list
            assert uninstall_call[0][0][1:] == ['-m', 'pip', 'uninstall', 'setuptools', '-y']



# SPDX-License-Identifier: MIT

import argparse
import os
import sys
import traceback
import warnings

from typing import Iterable, List, Optional, Sequence, TextIO, Type, Union

import build

from build import BuildBackendException, BuildException, ConfigSettings, ProjectBuilder
from build.env import IsolatedEnvBuilder


__all__ = ['build', 'main', 'main_parser']


def _showwarning(message, category, filename, lineno, file=None, line=None):  # pragma: no cover
    # type: (Union[Warning, str], Type[Warning], str, int, Optional[TextIO], Optional[str]) -> None
    prefix = 'WARNING'
    if sys.stdout.isatty():
        prefix = '\33[93m' + prefix + '\33[0m'
    print('{} {}'.format(prefix, str(message)))


warnings.showwarning = _showwarning


def _error(msg, code=1):  # type: (str, int) -> None  # pragma: no cover
    """
    Print an error message and exit. Will color the output when writing to a TTY.

    :param msg: Error message
    :param code: Error code
    """
    prefix = 'ERROR'
    if sys.stdout.isatty():
        prefix = '\33[91m' + prefix + '\33[0m'
    print('{} {}'.format(prefix, msg))
    exit(code)


def _format_dep_chain(dep_chain):  # type: (Sequence[str]) -> str
    return ' -> '.join(dep.partition(';')[0].strip() for dep in dep_chain)


def _build_in_isolated_env(builder, outdir, distributions, config_settings):
    # type: (ProjectBuilder, str, List[str], ConfigSettings) -> None
    for distribution in distributions:
        with IsolatedEnvBuilder() as env:
            builder.python_executable = env.executable
            builder.scripts_dir = env.scripts_dir
            # first install the build dependencies
            env.install(builder.build_dependencies)
            # then get the extra required dependencies from the backend (which was installed in the call above :P)
            env.install(builder.get_dependencies(distribution))
            builder.build(distribution, outdir, config_settings)


def _build_in_current_env(builder, outdir, distributions, config_settings, skip_dependencies=False):
    # type: (ProjectBuilder, str, List[str], ConfigSettings, bool) -> None
    for dist in distributions:
        if not skip_dependencies:
            missing = builder.check_dependencies(dist)
            if missing:
                _error(
                    'Missing dependencies:'
                    + ''.join('\n\t' + dep for deps in missing for dep in (deps[0], _format_dep_chain(deps[1:])) if dep)
                )

        builder.build(dist, outdir, config_settings)


def build_package(srcdir, outdir, distributions, config_settings=None, isolation=True, skip_dependencies=False):
    # type: (str, str, List[str], Optional[ConfigSettings], bool, bool) -> None
    """
    Run the build process.

    :param srcdir: Source directory
    :param outdir: Output directory
    :param distributions: Distributions to build (sdist and/or wheel)
    :param config_settings: Configuration settings to be passed to the backend
    :param isolation: Isolate the build in a separate environment
    :param skip_dependencies: Do not perform the dependency check
    """
    if not config_settings:
        config_settings = {}

    try:
        builder = ProjectBuilder(srcdir)
        if isolation:
            _build_in_isolated_env(builder, outdir, distributions, config_settings)
        else:
            _build_in_current_env(builder, outdir, distributions, config_settings, skip_dependencies)
    except BuildException as e:
        _error(str(e))
    except BuildBackendException as e:
        if sys.version_info >= (3, 5):
            print(traceback.format_exc(-1))
        else:
            print(traceback.format_exc())
        _error(str(e))


def main_parser():  # type: () -> argparse.ArgumentParser
    """
    Construct the main parser.
    """
    # mypy does not recognize module.__path__
    # https://github.com/python/mypy/issues/1422
    paths = build.__path__  # type: Iterable[Optional[str]]  # type: ignore
    parser = argparse.ArgumentParser()
    parser.add_argument(
        'srcdir',
        type=str,
        nargs='?',
        default=os.getcwd(),
        help='source directory (defaults to current directory)',
    )
    parser.add_argument(
        '--version',
        '-V',
        action='version',
        version='build {} ({})'.format(build.__version__, ', '.join(path for path in paths if path)),
    )
    parser.add_argument(
        '--sdist',
        '-s',
        action='store_true',
        help='build a source distribution',
    )
    parser.add_argument(
        '--wheel',
        '-w',
        action='store_true',
        help='build a wheel',
    )
    parser.add_argument(
        '--outdir',
        '-o',
        type=str,
        help='output directory (defaults to {{srcdir}}{sep}dist)'.format(sep=os.sep),
    )
    parser.add_argument(
        '--skip-dependencies',
        '-x',
        action='store_true',
        help='do not check that build dependencies are installed',
    )
    parser.add_argument(
        '--no-isolation',
        '-n',
        action='store_true',
        help='do not isolate the build in a virtual environment',
    )
    parser.add_argument(
        '--config-setting',
        '-C',
        action='append',
        help='pass options to the backend.  options which begin with a hyphen must be in the form of '
        '"--config-setting=--opt(=value)" or "-C--opt(=value)"',
    )
    return parser


def main(cli_args, prog=None):  # type: (List[str], Optional[str]) -> None
    """
    Parse the CLI arguments and invoke the build process.

    :param cli_args: CLI arguments
    :param prog: Program name to show in help text
    """
    parser = main_parser()
    if prog:
        parser.prog = prog
    args = parser.parse_args(cli_args)

    distributions = []
    config_settings = {}

    if args.config_setting:
        for arg in args.config_setting:
            setting, _, value = arg.partition('=')
            if setting not in config_settings:
                config_settings[setting] = value
            else:
                if not isinstance(config_settings[setting], list):
                    config_settings[setting] = [config_settings[setting]]

                config_settings[setting].append(value)

    if args.sdist:
        distributions.append('sdist')
    if args.wheel:
        distributions.append('wheel')

    # default targets
    if not distributions:
        distributions = ['sdist', 'wheel']

    # outdir is relative to srcdir only if omitted.
    outdir = os.path.join(args.srcdir, 'dist') if args.outdir is None else args.outdir

    build_package(args.srcdir, outdir, distributions, config_settings, not args.no_isolation, args.skip_dependencies)


def entrypoint():  # type: () -> None
    main(sys.argv[1:])


if __name__ == '__main__':  # pragma: no cover
    main(sys.argv[1:], 'python -m build')



import abc
import sys

from typing import TYPE_CHECKING, Callable, TypeVar


_T = TypeVar('_T')


def add_metaclass(metaclass):  # type: (type) -> Callable[[_T], _T]
    """Class decorator for creating a class with a metaclass (borrowed from six code)."""

    def wrapper(cls):  # type: (_T) -> _T
        orig_vars = cls.__dict__.copy()
        slots = orig_vars.get('__slots__')
        if slots is not None:
            if isinstance(slots, str):  # pragma: no cover
                slots = [slots]  # pragma: no cover
            for slots_var in slots:  # pragma: no cover
                orig_vars.pop(slots_var)  # pragma: no cover
        orig_vars.pop('__dict__', None)
        orig_vars.pop('__weakref__', None)
        if hasattr(cls, '__qualname__'):
            orig_vars['__qualname__'] = cls.__qualname__  # type: ignore
        return metaclass(cls.__name__, cls.__bases__, orig_vars)  # type: ignore

    return wrapper


if sys.version_info[0] == 2:
    abstractproperty = abc.abstractproperty
else:
    if TYPE_CHECKING:  # pragma: no cover
        abstractproperty = property  # pragma: no cover
    else:

        def abstractproperty(func):
            return property(abc.abstractmethod(func))


__all__ = (
    'abstractproperty',
    'add_metaclass',
)



# SPDX-License-Identifier: MIT

"""
build - A simple, correct PEP 517 package builder
"""
__version__ = '0.3.1'

import contextlib
import difflib
import io
import os
import sys
import warnings

from collections import OrderedDict
from typing import AbstractSet, Any, Callable, Dict, Iterator, Mapping, Optional, Sequence, Set, Text, Tuple, Union

import pep517.wrappers
import toml
import toml.decoder


if sys.version_info < (3,):
    FileNotFoundError = IOError
    PermissionError = OSError


ConfigSettings = Mapping[str, Union[str, Sequence[str]]]


_DEFAULT_BACKEND = {
    'build-backend': 'setuptools.build_meta:__legacy__',
    'requires': ['setuptools >= 40.8.0', 'wheel'],
}


class BuildException(Exception):
    """
    Exception raised by ProjectBuilder
    """


class BuildBackendException(Exception):
    """
    Exception raised when the backend fails
    """

    def __init__(self, exception):  # type: (Exception) -> None
        super(BuildBackendException, self).__init__()
        self.exception = exception  # type: Exception

    def __str__(self):  # type: () -> str
        return 'Backend operation failed: {!r}'.format(self.exception)


class TypoWarning(Warning):
    """
    Warning raised when a potential typo is found
    """


def check_dependency(req_string, ancestral_req_strings=(), parent_extras=frozenset()):
    # type: (str, Tuple[str, ...], AbstractSet[str]) -> Iterator[Tuple[str, ...]]
    """
    Verify that a dependency and all of its dependencies are met.

    :param req_string: Requirement string
    :param parent_extras: Extras (eg. "test" in myproject[test])
    :yields: Unmet dependencies
    """
    import packaging.requirements

    if sys.version_info >= (3, 8):
        import importlib.metadata as importlib_metadata
    else:
        import importlib_metadata

    req = packaging.requirements.Requirement(req_string)

    if req.marker:
        extras = frozenset(('',)).union(parent_extras)
        # a requirement can have multiple extras but ``evaluate`` can
        # only check one at a time.
        if all(not req.marker.evaluate(environment={'extra': e}) for e in extras):
            # if the marker conditions are not met, we pretend that the
            # dependency is satisfied.
            return

    try:
        dist = importlib_metadata.distribution(req.name)
    except importlib_metadata.PackageNotFoundError:
        # dependency is not installed in the environment.
        yield ancestral_req_strings + (req_string,)
    else:
        if req.specifier and not req.specifier.contains(dist.version, prereleases=True):
            # the installed version is incompatible.
            yield ancestral_req_strings + (req_string,)
        elif dist.requires:
            for other_req_string in dist.requires:
                for unmet_req in check_dependency(other_req_string, ancestral_req_strings + (req_string,), req.extras):
                    # a transitive dependency is not satisfied.
                    yield unmet_req


def _find_typo(dictionary, expected):  # type: (Mapping[str, str], str) -> None
    if expected not in dictionary:
        for obj in dictionary:
            if difflib.SequenceMatcher(None, expected, obj).ratio() >= 0.8:
                warnings.warn(
                    "Found '{}' in pyproject.toml, did you mean '{}'?".format(obj, expected),
                    TypoWarning,
                )


@contextlib.contextmanager
def _working_directory(path):  # type: (str) -> Iterator[None]
    current = os.getcwd()

    os.chdir(path)

    try:
        yield
    finally:
        os.chdir(current)


class ProjectBuilder(object):
    """
    The PEP 517 consumer API.
    """

    def __init__(self, srcdir, python_executable=sys.executable, scripts_dir=None):
        # type: (str, Union[bytes, Text], Optional[Union[bytes, Text]]) -> None
        """
        :param srcdir: The source directory
        :param scripts_dir: The location of the scripts dir (defaults to the folder where the python executable lives)
        :param python_executable: The python executable where the backend lives
        """
        self.srcdir = os.path.abspath(srcdir)  # type: str

        spec_file = os.path.join(srcdir, 'pyproject.toml')

        try:
            with io.open(spec_file, encoding='UTF-8') as f:
                spec = toml.load(f)
        except FileNotFoundError:
            spec = {}
        except PermissionError as e:
            raise BuildException("{}: '{}' ".format(e.strerror, e.filename))
        except toml.decoder.TomlDecodeError as e:
            raise BuildException('Failed to parse {}: {} '.format(spec_file, e))

        build_system = spec.get('build-system')
        # if pyproject.toml is missing (per PEP 517) or [build-system] is missing (per PEP 518),
        # use default values.
        if build_system is None:
            _find_typo(spec, 'build-system')
            build_system = _DEFAULT_BACKEND
        # if [build-system] is present, it must have a ``requires`` field (per PEP 518).
        elif 'requires' not in build_system:
            _find_typo(build_system, 'requires')
            raise BuildException("Missing 'build-system.requires' in {}".format(spec_file))
        # if ``build-backend`` is missing, inject the legacy setuptools backend
        # but leave ``requires`` alone to emulate pip.
        elif 'build-backend' not in build_system:
            _find_typo(build_system, 'build-backend')
            build_system['build-backend'] = _DEFAULT_BACKEND['build-backend']

        self._build_system = build_system
        self._backend = self._build_system['build-backend']
        self._scripts_dir = scripts_dir
        self._hook = pep517.wrappers.Pep517HookCaller(
            self.srcdir,
            self._backend,
            backend_path=self._build_system.get('backend-path'),
            python_executable=python_executable,
            runner=self._runner,
        )

    def _runner(self, cmd, cwd=None, extra_environ=None):
        # type: (Sequence[str], Optional[Union[bytes, Text]], Optional[Dict[str, str]]) -> None
        # if script dir is specified must be inserted at the start of PATH (avoid duplicate path while doing so)
        if self.scripts_dir is not None:
            paths = OrderedDict()  # type:  Dict[str, None]
            paths[str(self.scripts_dir)] = None
            if 'PATH' in os.environ:
                paths.update((i, None) for i in os.environ['PATH'].split(os.pathsep))
            extra_environ = {} if extra_environ is None else extra_environ
            extra_environ['PATH'] = os.pathsep.join(paths)
        pep517.default_subprocess_runner(cmd, cwd, extra_environ)

    @property
    def python_executable(self):  # type: () -> Union[bytes, Text]
        """
        The Python executable used to invoke the backend.
        """
        # make mypy happy
        exe = self._hook.python_executable  # type: Union[bytes, Text]
        return exe

    @python_executable.setter
    def python_executable(self, value):  # type: (Union[bytes, Text]) -> None
        self._hook.python_executable = value

    @property
    def scripts_dir(self):  # type: () -> Union[None, bytes, Text]
        """
        The folder where the scripts are stored for the python executable.
        """
        return self._scripts_dir

    @scripts_dir.setter
    def scripts_dir(self, value):  # type: (Union[None, bytes, Text]) -> None
        self._scripts_dir = value

    @property
    def build_dependencies(self):  # type: () -> Set[str]
        """
        The dependencies defined in the ``pyproject.toml``'s
        ``build-system.requires`` field or the default build dependencies
        if ``pyproject.toml`` is missing or ``build-system`` is undefined.
        """
        return set(self._build_system['requires'])

    def get_dependencies(self, distribution, config_settings=None):  # type: (str, Optional[ConfigSettings]) -> Set[str]
        """
        Return the dependencies defined by the backend in addition to
        :attr:`build_dependencies` for a given distribution.

        :param distribution: Distribution to get the dependencies of
            (``sdist`` or ``wheel``)
        :param config_settings: Config settings for the build backend
        """
        get_requires = getattr(self._hook, 'get_requires_for_build_{}'.format(distribution))

        with _working_directory(self.srcdir):
            try:
                return set(get_requires(config_settings))
            except pep517.wrappers.BackendUnavailable:
                raise BuildException("Backend '{}' is not available.".format(self._backend))
            except Exception as e:
                raise BuildBackendException(e)

    def check_dependencies(self, distribution, config_settings=None):
        # type: (str, Optional[ConfigSettings]) -> Set[Tuple[str, ...]]
        """
        Return the dependencies which are not satisfied from the combined set of
        :attr:`build_dependencies` and :meth:`get_dependencies` for a given
        distribution.

        :param distribution: Distribution to check (``sdist`` or ``wheel``)
        :param config_settings: Config settings for the build backend
        :returns: Set of variable-length unmet dependency tuples
        """
        dependencies = self.get_dependencies(distribution, config_settings).union(self.build_dependencies)
        return {u for d in dependencies for u in check_dependency(d)}

    def prepare(self, distribution, output_directory, config_settings=None):
        # type: (str, str, Optional[ConfigSettings]) -> Optional[str]
        """
        Prepare metadata for a distribution.

        :param distribution: Distribution to build (must be ``wheel``)
        :param output_directory: Directory to put the prepared metadata in
        :param config_settings: Config settings for the build backend
        :returns: The full path to the prepared metadata directory
        """
        prepare = getattr(self._hook, 'prepare_metadata_for_build_{}'.format(distribution))
        try:
            return self._call_backend(prepare, output_directory, config_settings, _allow_fallback=False)
        except BuildBackendException as exception:
            if isinstance(exception.exception, pep517.wrappers.HookMissing):
                return None
            raise

    def build(self, distribution, output_directory, config_settings=None, metadata_directory=None):
        # type: (str, str, Optional[ConfigSettings], Optional[str]) -> str
        """
        Build a distribution.

        :param distribution: Distribution to build (``sdist`` or ``wheel``)
        :param output_directory: Directory to put the built distribution in
        :param config_settings: Config settings for the build backend
        :param metadata_directory: If provided, should be the return value of a
            previous ``prepare`` call on the same ``distribution`` kind
        :returns: The full path to the built distribution
        """
        build = getattr(self._hook, 'build_{}'.format(distribution))
        kwargs = {} if metadata_directory is None else {'metadata_directory': metadata_directory}
        return self._call_backend(build, output_directory, config_settings, **kwargs)

    def _call_backend(self, callback, outdir, config_settings=None, **kwargs):
        # type: (Callable[...,str], str, Optional[ConfigSettings], Any) -> str
        outdir = os.path.abspath(outdir)

        if os.path.exists(outdir):
            if not os.path.isdir(outdir):
                raise BuildException("Build path '{}' exists and is not a directory".format(outdir))
        else:
            os.mkdir(outdir)

        with _working_directory(self.srcdir):
            try:
                basename = callback(outdir, config_settings, **kwargs)  # type: str
                return os.path.join(outdir, basename)
            except pep517.wrappers.BackendUnavailable:
                raise BuildException("Backend '{}' is not available.".format(self._backend))
            except Exception as exception:
                raise BuildBackendException(exception)


__all__ = (
    '__version__',
    'ConfigSettings',
    'BuildException',
    'BuildBackendException',
    'TypoWarning',
    'check_dependency',
    'ProjectBuilder',
)
